import { getAuthHeaders } from '@/api/auth'

const BASE_URL = import.meta.env.VITE_API_URL + '/classroom-compliance'

export interface Class {
  id: number
  number: number
  schoolYear: string
}

export interface Student {
  id: number
  name: string
  classId: number
  deskNumber: number
  removed: boolean
}

export interface EntryResponseItemPhone {
  compliant: boolean
}

export interface EntryResponseItemBehavior {
  rating: number
  comment?: string
}

export interface EntryResponseItem {
  id: number
  date: string
  createdAt: string
  absent: boolean
  phone?: EntryResponseItemPhone
  behavior?: EntryResponseItemBehavior
}

export interface EntryResponseStudent {
  id: number
  name: string
  deskNumber: number
  removed: boolean
  entries: Record<string, EntryResponseItem | null>
}

export interface EntriesResponse {
  students: EntryResponseStudent[]
  dates: string[]
}

function getClassPath(classId: number): string {
  return BASE_URL + '/classes/' + classId
}

export async function createClass(
  auth: string,
  number: number,
  schoolYear: string,
): Promise<Class> {
  const response = await fetch(BASE_URL + '/classes', {
    method: 'POST',
    headers: getAuthHeaders(auth),
    body: JSON.stringify({ number: number, schoolYear: schoolYear }),
  })
  return (await response.json()) as Class
}

export async function getClasses(auth: string): Promise<Class[]> {
  const response = await fetch(BASE_URL + '/classes', {
    headers: getAuthHeaders(auth),
  })
  return (await response.json()) as Class[]
}

export async function getClass(auth: string, id: number): Promise<Class> {
  const response = await fetch(getClassPath(id), {
    headers: getAuthHeaders(auth),
  })
  return (await response.json()) as Class
}

export async function deleteClass(auth: string, classId: number): Promise<void> {
  const response = await fetch(getClassPath(classId), {
    method: 'DELETE',
    headers: getAuthHeaders(auth),
  })
  if (!response.ok) {
    throw new Error('Failed to delete class.')
  }
}

export async function getStudents(auth: string, classId: number): Promise<Student[]> {
  const response = await fetch(getClassPath(classId) + '/students', {
    headers: getAuthHeaders(auth),
  })
  return (await response.json()) as Student[]
}

export async function getEntries(
  auth: string,
  classId: number,
  fromDate?: Date,
  toDate?: Date,
): Promise<EntriesResponse> {
  const params = new URLSearchParams()
  if (fromDate) {
    params.append('from', fromDate.toISOString().substring(0, 10))
  }
  if (toDate) {
    params.append('to', toDate.toISOString().substring(0, 10))
  }
  const response = await fetch(getClassPath(classId) + '/entries?' + params.toString(), {
    headers: getAuthHeaders(auth),
  })
  return (await response.json()) as EntriesResponse
}
